import os
import sys
from vex.run import run
from vex import exceptions


PYDOC_SCRIPT = """#!/usr/bin/env python
from pydoc import cli
cli()
""".encode('ascii')


PYDOC_BATCH = """
@python -m pydoc %*
""".encode('ascii')


def handle_make(environ, options, make_path):
    if os.path.exists(make_path):
        # Can't ignore existing virtualenv happily because existing one
        # might have different parameters and --make implies nonexistent
        raise exceptions.VirtualenvAlreadyMade(make_path)
    ve_base = os.path.dirname(make_path)
    if not os.path.exists(ve_base):
        os.mkdir(ve_base)
    elif not os.path.isdir(ve_base):
        raise exceptions.VirtualenvNotMade(
            "could not make virtualenv: "
            "{0!r} already exists but is not a directory. "
            "Choose a different virtualenvs path using ~/.vexrc "
            "or $WORKON_HOME, or remove the existing file; "
            "then rerun your vex --make command.".format(ve_base)
        )
    # TODO: virtualenv is usually not on PATH for Windows,
    # but finding it is a terrible issue.
    if os.name == 'nt' and not os.environ.get('VIRTUAL_ENV', ''):
        ve = os.path.join(
            os.path.dirname(sys.executable),
            'Scripts',
            'virtualenv'
        )
    else:
        ve = 'virtualenv'
    args = [ve, make_path]
    if options.python:
        args += ['--python', options.python]
    if options.site_packages:
        args += ['--site-packages']
    if options.always_copy:
        args+= ['--always-copy']
    returncode = run(args, env=environ, cwd=ve_base)
    if returncode != 0:
        raise exceptions.VirtualenvNotMade("error creating virtualenv")
    if os.name != 'nt':
        pydoc_path = os.path.join(make_path, 'bin', 'pydoc')
        with open(pydoc_path, 'wb') as out:
            out.write(PYDOC_SCRIPT)
        perms = os.stat(pydoc_path).st_mode
        os.chmod(pydoc_path, perms | 0o0111)
    else:
        pydoc_path = os.path.join(make_path, 'Scripts', 'pydoc.bat')
        with open(pydoc_path, 'wb') as out:
            out.write(PYDOC_BATCH)



import sys


class InvalidArgument(Exception):
    """Raised by anything under main() to propagate errors to user.
    """
    def __init__(self, message):
        self.message = message
        Exception.__init__(self, message)


class NoVirtualenvName(InvalidArgument):
    """No virtualenv name was given (insufficient arguments).
    """
    pass


class NoVirtualenvsDirectory(InvalidArgument):
    """There is no directory to find named virtualenvs in.
    """
    pass


class OtherShell(InvalidArgument):
    """The given argument to --shell-config is not recognized.
    """
    pass


class UnknownArguments(InvalidArgument):
    """Unknown arguments were given on the command line.

    This is a byproduct of having to use parse_known_args.
    """
    pass


class InvalidVexrc(InvalidArgument):
    """config file specified or required but absent or unparseable.
    """
    pass


class InvalidVirtualenv(InvalidArgument):
    """No usable virtualenv was found.
    """
    pass


class InvalidCommand(InvalidArgument):
    """No runnable command was found.
    """
    pass


class InvalidCwd(InvalidArgument):
    """cwd specified or required but unusable.
    """
    pass


class BadConfig(InvalidArgument):
    """raised to halt on fatal conditions on the way to run.
    """
    pass


class VirtualenvAlreadyMade(InvalidArgument):
    """could not make virtualenv as one already existed.
    """
    pass


class VirtualenvNotMade(InvalidArgument):
    """could not make virtualenv.
    """
    pass


class VirtualenvNotRemoved(InvalidArgument):
    """raised when virtualenv could not be removed.
    """
    pass


if sys.version_info > (3, 3):
    CommandNotFoundError = FileNotFoundError
else:
    CommandNotFoundError = OSError



"""Config file processing (.vexrc).
"""
import os
import sys
import re
import shlex
import platform
from collections import OrderedDict

_IDENTIFIER_PATTERN = '[a-zA-Z][_a-zA-Z0-9]*'
_SQUOTE_RE = re.compile(r"'([^']*)'\Z")  # NO squotes inside
_DQUOTE_RE = re.compile(r'"([^"]*)"\Z')  # NO dquotes inside
_HEADING_RE = re.compile(
    r'^(' + _IDENTIFIER_PATTERN + r'):[ \t\n\r]*\Z')
_VAR_RE = re.compile(
    r'[ \t]*(' + _IDENTIFIER_PATTERN + r') *= *(.*)[ \t\n\r]*$')


if sys.version_info < (3, 3):
    FileNotFoundError = IOError


class InvalidConfigError(Exception):
    """Raised when there is an error during a .vexrc file parse.
    """
    def __init__(self, filename, errors):
        Exception.__init__(self)
        self.filename = filename
        self.errors = errors

    def __str__(self):
        return "errors in {0!r}, lines {1!r}".format(
            self.filename,
            list(tup[0] for tup in self.errors)
        )


class Vexrc(object):
    """Parsed representation of a .vexrc config file.
    """
    default_heading = "root"
    default_encoding = "utf-8"

    def __init__(self):
        self.encoding = self.default_encoding
        self.headings = OrderedDict()
        self.headings[self.default_heading] = OrderedDict()
        self.headings['env'] = OrderedDict()

    def __getitem__(self, key):
        return self.headings.get(key)

    @classmethod
    def from_file(cls, path, environ):
        """Make a Vexrc instance from given file in given environ.
        """
        instance = cls()
        instance.read(path, environ)
        return instance

    def read(self, path, environ):
        """Read data from file into this vexrc instance.
        """
        try:
            inp = open(path, 'rb')
        except FileNotFoundError as error:
            if error.errno != 2:
                raise
            return None
        parsing = parse_vexrc(inp, environ)
        for heading, key, value in parsing:
            heading = self.default_heading if heading is None else heading
            if heading not in self.headings:
                self.headings[heading] = OrderedDict()
            self.headings[heading][key] = value
        parsing.close()

    def get_ve_base(self, environ):
        """Find a directory to look for virtualenvs in.
        """
        # set ve_base to a path we can look for virtualenvs:
        # 1. .vexrc
        # 2. WORKON_HOME (as defined for virtualenvwrapper's benefit)
        # 3. $HOME/.virtualenvs
        # (unless we got --path, then we don't need it)
        ve_base_value = self.headings[self.default_heading].get('virtualenvs')
        if ve_base_value:
            ve_base = os.path.expanduser(ve_base_value)
        else:
            ve_base = environ.get('WORKON_HOME', '')
        if not ve_base:
            # On Cygwin os.name == 'posix' and we want $HOME.
            if platform.system() == 'Windows' and os.name == 'nt':
                _win_drive = environ.get('HOMEDRIVE')
                home = environ.get('HOMEPATH', '')
                if home:
                    home = os.path.join(_win_drive, home)
            else:
                home = environ.get('HOME', '')
            if not home:
                home = os.path.expanduser('~')
            if not home:
                return ''
            ve_base = os.path.join(home, '.virtualenvs')
        # pass through invalid paths so messages can be generated
        # if not os.path.exists(ve_base) or os.path.isfile(ve_base):
            # return ''
        return ve_base or ''

    def get_shell(self, environ):
        """Find a command to run.
        """
        command = self.headings[self.default_heading].get('shell')
        if not command and os.name != 'nt':
            command = environ.get('SHELL', '')
        command = shlex.split(command) if command else None
        return command


def extract_heading(line):
    """Return heading in given line or None if it's not a heading.
    """
    match = _HEADING_RE.match(line)
    if match:
        return match.group(1)
    return None


def extract_key_value(line, environ):
    """Return key, value from given line if present, else return None.
    """
    segments = line.split("=", 1)
    if len(segments) < 2:
        return None
    key, value = segments
    # foo passes through as-is (with spaces stripped)
    # '{foo}' passes through literally
    # "{foo}" substitutes from environ's foo
    value = value.strip()
    if value[0] == "'" and _SQUOTE_RE.match(value):
        value = value[1:-1]
    elif value[0] == '"' and _DQUOTE_RE.match(value):
        template = value[1:-1]
        value = template.format(**environ)
    key = key.strip()
    value = value.strip()
    return key, value


def parse_vexrc(inp, environ):
    """Iterator yielding key/value pairs from given stream.

    yields tuples of heading, key, value.
    """
    heading = None
    errors = []
    with inp:
        for line_number, line in enumerate(inp):
            line = line.decode("utf-8")
            if not line.strip():
                continue
            extracted_heading = extract_heading(line)
            if extracted_heading is not None:
                heading = extracted_heading
                continue
            kv_tuple = extract_key_value(line, environ)
            if kv_tuple is None:
                errors.append((line_number, line))
                continue
            try:
                yield heading, kv_tuple[0], kv_tuple[1]
            except GeneratorExit:
                break
    if errors:
        raise InvalidConfigError(inp.name, errors)



import argparse
from vex import exceptions


def make_arg_parser():
    """Return a standard ArgumentParser object.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        usage="vex [OPTIONS] VIRTUALENV_NAME COMMAND_TO_RUN ...",
    )

    make = parser.add_argument_group(title='To make a new virtualenv')
    make.add_argument(
        '-m', '--make',
        action="store_true",
        help="make named virtualenv before running command"
    )
    make.add_argument(
        '--python',
        help="specify which python for virtualenv to be made",
        action="store",
        default="python",
    )
    make.add_argument(
        '--site-packages',
        help="allow site package imports from new virtualenv",
        action="store_true",
    )
    make.add_argument(
        '--always-copy',
        help="use copies instead of symlinks in new virtualenv",
        action="store_true",
    )

    remove = parser.add_argument_group(title='To remove a virtualenv')
    remove.add_argument(
        '-r', '--remove',
        action="store_true",
        help="remove the named virtualenv after running command"
    )

    parser.add_argument(
        "--path",
        metavar="DIR",
        help="absolute path to virtualenv to use",
        action="store"
    )
    parser.add_argument(
        '--cwd',
        metavar="DIR",
        action="store",
        default='.',
        help="path to run command in (default: '.' aka $PWD)",
    )
    parser.add_argument(
        "--config",
        metavar="FILE",
        default=None,
        action="store",
        help="path to config file to read (default: '~/.vexrc')"
    )
    parser.add_argument(
        '--shell-config',
        metavar="SHELL",
        dest="shell_to_configure",
        action="store",
        default=None,
        help="print optional config for the specified shell"
    )
    parser.add_argument(
        "rest",
        nargs=argparse.REMAINDER,
        help=argparse.SUPPRESS)

    return parser



def get_options(argv):
    """Called to parse the given list as command-line arguments.

    :returns:
        an options object as returned by argparse.
    """
    arg_parser = make_arg_parser()
    options, unknown = arg_parser.parse_known_args(argv)
    if unknown:
        arg_parser.print_help()
        raise exceptions.UnknownArguments(
            "unknown args: {0!r}".format(unknown))
    options.print_help = arg_parser.print_help
    return options



"""Main command-line entry-point and any code tightly coupled to it.
"""
import sys
import os
from vex import config
from vex.options import get_options
from vex.run import get_environ, run
from vex.shell_config import handle_shell_config
from vex.make import handle_make
from vex.remove import handle_remove
from vex import exceptions


def get_vexrc(options, environ):
    """Get a representation of the contents of the config file.

    :returns:
        a Vexrc instance.
    """
    # Complain if user specified nonexistent file with --config.
    # But we don't want to complain just because ~/.vexrc doesn't exist.
    if options.config and not os.path.exists(options.config):
        raise exceptions.InvalidVexrc("nonexistent config: {0!r}".format(options.config))
    filename = options.config or os.path.expanduser('~/.vexrc')
    vexrc = config.Vexrc.from_file(filename, environ)
    return vexrc


def get_cwd(options):
    """Discover what directory the command should run in.
    """
    if not options.cwd:
        return None
    if not os.path.exists(options.cwd):
        raise exceptions.InvalidCwd(
            "can't --cwd to invalid path {0!r}".format(options.cwd))
    return options.cwd


def get_virtualenv_name(options):
    if options.path:
        return os.path.dirname(options.path)
    else:
        ve_name = options.rest.pop(0) if options.rest else ''
    if not ve_name:
        raise exceptions.NoVirtualenvName(
            "could not find a virtualenv name in the command line."
        )
    return ve_name


def get_virtualenv_path(ve_base, ve_name):
    """Check a virtualenv path, raising exceptions to explain problems.
    """
    if not ve_base:
        raise exceptions.NoVirtualenvsDirectory(
            "could not figure out a virtualenvs directory. "
            "make sure $HOME is set, or $WORKON_HOME,"
            " or set virtualenvs=something in your .vexrc")

    # Using this requires get_ve_base to pass through nonexistent dirs
    if not os.path.exists(ve_base):
        message = (
            "virtualenvs directory {0!r} not found. "
            "Create it or use vex --make to get started."
        ).format(ve_base)
        raise exceptions.NoVirtualenvsDirectory(message)

    if not ve_name:
        raise exceptions.InvalidVirtualenv("no virtualenv name")

    # n.b.: if ve_name is absolute, ve_base is discarded by os.path.join,
    # and an absolute path will be accepted as first arg.
    # So we check if they gave an absolute path as ve_name.
    # But we don't want this error if $PWD == $WORKON_HOME,
    # in which case 'foo' is a valid relative path to virtualenv foo.
    ve_path = os.path.join(ve_base, ve_name)
    if ve_path == ve_name and os.path.basename(ve_name) != ve_name:
        raise exceptions.InvalidVirtualenv(
            'To run in a virtualenv by its path, '
            'use "vex --path {0}"'.format(ve_path))

    ve_path = os.path.abspath(ve_path)
    if not os.path.exists(ve_path):
        raise exceptions.InvalidVirtualenv(
            "no virtualenv found at {0!r}.".format(ve_path))
    return ve_path


def get_command(options, vexrc, environ):
    """Get a command to run.

    :returns:
        a list of strings representing a command to be passed to Popen.
    """
    command = options.rest
    if not command:
        command = vexrc.get_shell(environ)
    if command and command[0].startswith('--'):
        raise exceptions.InvalidCommand(
            "don't put flags like '%s' after the virtualenv name."
            % command[0])
    if not command:
        raise exceptions.InvalidCommand("no command given")
    return command


def _main(environ, argv):
    """Logic for main(), with less direct system interaction.

    Routines called here raise InvalidArgument with messages that
    should be delivered on stderr, to be caught by main.
    """
    options = get_options(argv)
    vexrc = get_vexrc(options, environ)
    # Handle --shell-config as soon as its arguments are available.
    if options.shell_to_configure:
        return handle_shell_config(options.shell_to_configure, vexrc, environ)
    # Do as much as possible before a possible make, so errors can raise
    # without leaving behind an unused virtualenv.
    # get_virtualenv_name is destructive and must happen before get_command
    cwd = get_cwd(options)
    ve_base = vexrc.get_ve_base(environ)
    ve_name = get_virtualenv_name(options)
    command = get_command(options, vexrc, environ)
    # Either we create ve_path, get it from options.path or find it
    # in ve_base.
    if options.make:
        if options.path:
            make_path = os.path.abspath(options.path)
        else:
            make_path = os.path.abspath(os.path.join(ve_base, ve_name))
        handle_make(environ, options, make_path)
        ve_path = make_path
    elif options.path:
        ve_path = os.path.abspath(options.path)
    else:
        try:
            ve_path = get_virtualenv_path(ve_base, ve_name)
        except exceptions.NoVirtualenvName:
            options.print_help()
            raise
    # get_environ has to wait until ve_path is defined, which might
    # be after a make; of course we can't run until we have env.
    env = get_environ(environ, vexrc['env'], ve_path)
    returncode = run(command, env=env, cwd=cwd)
    if options.remove:
        handle_remove(ve_path)
    if returncode is None:
        raise exceptions.InvalidCommand(
            "command not found: {0!r}".format(command[0]))
    return returncode


def main():
    """The main command-line entry point, with system interactions.
    """
    argv = sys.argv[1:]
    returncode = 1
    try:
        returncode = _main(os.environ, argv)
    except exceptions.InvalidArgument as error:
        if error.message:
            sys.stderr.write("Error: " + error.message + '\n')
        else:
            raise
    sys.exit(returncode)
